/*! ********************************************
 * GrayBoxSound.java
 * Copyright 2019 masahiro takamoto (MotChan)
 * Distributed under the MIT license. 
 * see -> sora/license.txt
 ******************************************** */
package sora.graybox;

import java.util.Timer;
import java.util.TimerTask;
import javax.sound.sampled.AudioFormat;
import javax.sound.sampled.AudioSystem;
import javax.sound.sampled.DataLine;
import javax.sound.sampled.LineEvent;
import javax.sound.sampled.LineListener;
import javax.sound.sampled.LineUnavailableException;
import javax.sound.sampled.SourceDataLine;

public final class GrayBoxSound implements LineListener
{
	@FunctionalInterface
	public interface Func { void apply(); }

	// object
	private Timer _timer = null;
	private GrayBox _grayblox = null;
	// AudioBuffer
	private int _audioBufferSize = 4096;
	private float[][] _audioBuffer = null;
	// sourceDataLine
	private SourceDataLine _sdLine = null;
	private int _bufferBytes = 0;
	private int _dataLineBytes = 0;
	private int _dlBytes2 = 0;
	private byte[] _bytes = null;
	private long _totalFrames = 0l;
	private long _frameCnt = 0l;
	private float _frameTime = 1000.0f / 44100.0f;
	// status
	private boolean _pause = true;
	private boolean _stopFlag = false;
	private Func _callBackStopProcess = null;
	// process
	public Func preProcess = null;
	public Func postProcess = null;

// buffer -----
	private void _clearAudioBuffer()
	{
		float[] _bufferL = _audioBuffer[0],
		        _bufferR = _audioBuffer[1];
		for (int i = 0; i < _audioBufferSize; i++)
		{
			_bufferL[i] = 0.0f;
			_bufferR[i] = 0.0f;
		}
	}
	public int getAudioBufferSize(){ return _audioBufferSize; }
	public int setAudioBufferSize(int size)
	{
		if (size == _audioBufferSize) return _audioBufferSize;
		if (_timer != null)return _audioBufferSize;
		int v = 16;
		for ( ; v < 65536; v <<= 1) if (v >= size) break;
		_audioBufferSize = v;
		_audioBuffer[0] = new float[v];
		_audioBuffer[1] = new float[v];
		_clearAudioBuffer();
		_grayblox.setProcessBlockSize(_audioBufferSize);
		return _audioBufferSize;
	}
// status ---
	public boolean isPlay() { return _timer != null; }
	public boolean isPause() { return _pause; }
// Main System ---
	public void initialize()
	{
		stop();
		_clearAudioBuffer();
		_grayblox.initialize();
	}
// Module ---
	public int getModuleSlots() { return _grayblox.getModuleSlots(); }
	public int setModuleSlots(int num) { return _grayblox.setModuleSlots(num); }
	public IGrayBoxModule getModule(int index) { return _grayblox.getModule(index); }
	public IGrayBoxModule attachModule(int index, IGrayBoxModule module) { return _grayblox.attachModule(index, module); }
	public IGrayBoxModule detachModule(int index) { return _grayblox.detachModule(index); }
	public void resetModules() { _grayblox.resetModules(); }
	public void initiModules() { _grayblox.initiModules(); }
	public boolean getBypassStatus(int index) { return _grayblox.getBypassStatus(index); }
	public void bypassProcess(int index, boolean bypass) { _grayblox.bypassProcess(index, bypass); }
// control ---
	public boolean play()
	{ 
		return play(44100.0f);
	}
	public SourceDataLine play(float sampleRate)
	{
		SourceDataLine _s;
		_s = play(new AudioFormat(AudioFormat.Encoding.PCM_FLOAT,    sampleRate, Byte.SIZE * 4, 2, 8, sampleRate, java.nio.ByteOrder.nativeOrder() == java.nio.ByteOrder.BIG_ENDIAN));
		if (_s != null)return _s;
		_s = play(new AudioFormat(AudioFormat.Encoding.PCM_SIGNED,   sampleRate, Byte.SIZE * 2, 2, 4, sampleRate, java.nio.ByteOrder.nativeOrder() == java.nio.ByteOrder.BIG_ENDIAN));
		if (_s != null)return _s;
		_s = play(new AudioFormat(AudioFormat.Encoding.PCM_UNSIGNED, sampleRate, Byte.SIZE * 2, 2, 4, sampleRate, java.nio.ByteOrder.nativeOrder() == java.nio.ByteOrder.BIG_ENDIAN));
		if (_s != null)return _s;
		return null;
	}
	public SourceDataLine play(AudioFormat afmt)
	{
		return play(new DataLine.Info(SourceDataLine.class, afmt));
	}
	public SourceDataLine play(DataLine.Info info)
	{
		if (_timer != null) return null;
		if (_stopFlag) return null;
		SourceDataLine _sdl;
        try { _sdl = (SourceDataLine) AudioSystem.getLine(info); }
		catch (LineUnavailableException e) { return null; }
		catch (IllegalArgumentException e) { return null; }
		catch (SecurityException e) { return null; }

		AudioFormat _afmt = _sdl.getFormat();
		if (_afmt.getChannels() != 2)return null;
		if (_afmt.getFrameRate() != _afmt.getSampleRate())return null;
		if (_afmt.getEncoding() == AudioFormat.Encoding.PCM_FLOAT)
		{
			int _frameSize = _afmt.getFrameSize();
			if      (_frameSize ==  8) _process = this::_process_float;
			else if (_frameSize == 16) _process = this::_process_double;
			else return null;
		}
		else if (_afmt.getEncoding() == AudioFormat.Encoding.PCM_SIGNED)
		{
			int _frameSize = _afmt.getFrameSize();
			if      (_frameSize == 2) _process = this::_process_s8;
			else if (_frameSize == 4) _process = this::_process_s16;
			else if (_frameSize == 6) _process = this::_process_s24;
			else if (_frameSize == 8) _process = this::_process_s32;
			else return null;
		}
		else if (_afmt.getEncoding() == AudioFormat.Encoding.PCM_UNSIGNED)
		{
			int _frameSize = _afmt.getFrameSize();
			if      (_frameSize == 2) _process = this::_process_u8;
			else if (_frameSize == 4) _process = this::_process_u16;
			else if (_frameSize == 6) _process = this::_process_u24;
			else if (_frameSize == 8) _process = this::_process_u32;
			else return null;
		}
		else return null;

		_totalFrames = 0l;
		_frameCnt = (long)_audioBufferSize;
		_bufferBytes = _audioBufferSize * _afmt.getChannels() * (_afmt.getSampleSizeInBits() / Byte.SIZE);
		_dataLineBytes = _bufferBytes * 4;
		_dlBytes2 = _bufferBytes * 2;
		_bytes = new byte[_bufferBytes];
		_sdl.flush();
        try { _sdl.open(_afmt, _dataLineBytes); }
		catch (LineUnavailableException e) { return null; }
		catch (SecurityException e) { return null; }

		_sdLine = _sdl;
		_sdl.addLineListener(this);
		_sdl.start();
		//
		if (_afmt.isBigEndian());
		// graybox
		_frameTime = 1000.0f / _afmt.getSampleRate();
		_grayblox.setSampleRate(_afmt.getSampleRate());
		_grayblox.startProcess();
		// Timer
		_pause = false;
		_timer = new Timer(true);
		_timer.schedule(new _BuffTimerTask(this), 0, 1);
		return _sdLine;
	}
	public void stop(){ stop(null); }
	public void stop(Func callback)
	{
		if (_timer == null) return;
		if (_stopFlag) return;
		_callBackStopProcess = callback;
		_stopFlag = true;
		// _timer.cancel();
		// _timer = null;
	}
	private void _stopProcess()
	{
		if (_timer != null)
		{
			_timer.cancel();
			_timer = null;
			if (_sdLine != null)
			{
				_sdLine.removeLineListener(this);
				_sdLine.flush();
				_sdLine.stop();
				_sdLine.close();
				_sdLine = null;
			}
			_grayblox.stopProcess();
			if (_callBackStopProcess != null)_callBackStopProcess.apply();
			_callBackStopProcess = null;
			_stopFlag = false;
		}
	}
	public void resume()
	{
		if (_timer == null) return;
		if (_pause)_grayblox.resumeProcess();
		_pause = false;
	}
	public void pause()
	{
		if (_timer == null) return;
		if (!_pause)_grayblox.suspendProcess();
		_pause = true;
	}

// interface - LineListener
	public void update(LineEvent event)
	{
		LineEvent.Type type = event.getType();
		if (type == LineEvent.Type.CLOSE || type == LineEvent.Type.STOP)
		{
			_stopProcess();
		}
	}
	// Constructor
	public GrayBoxSound(){ this(4096, 4); }
	public GrayBoxSound(int audioBufferSize){ this(audioBufferSize, 4); }
	public GrayBoxSound(int audioBufferSize, int modules)
	{
		int v = 16;
		for ( ; v < 65536; v <<= 1) if (v >= audioBufferSize) break;
		_audioBufferSize = v;
		_audioBuffer = new float[2][];
		_audioBuffer[0] = new float[v];
		_audioBuffer[1] = new float[v];
		_clearAudioBuffer();
		_grayblox = new GrayBox(v, 44100.0f, 2, modules);
	}
// Process ---
	private Func _process = null;

	private void _process_float()
	{
		while ((_dataLineBytes - _sdLine.available()) <= _dlBytes2)
		{
			if (preProcess != null)preProcess.apply();
			if (_pause)for (int i = 0; i < _bufferBytes; i++)_bytes[i] = 0;
			else 
			{
				_clearAudioBuffer();
				_grayblox.audioLatency((_totalFrames - _sdLine.getLongFramePosition()) * _frameTime);
				_grayblox.process(_audioBuffer, 0, _audioBufferSize);
				float[] _bufferL = _audioBuffer[0],  _bufferR = _audioBuffer[1];
				int _rs0 = 0, _rs1 = 8, _rs2 = 16, _rs3 = 24;
				if (_sdLine.getFormat().isBigEndian()){_rs0 = 24; _rs1 = 16; _rs2 = 8; _rs3 = 0; }
				for (int i = 0, p = 0; i < _audioBufferSize; i++)
				{
					int B = Float.floatToIntBits(_bufferL[i]);
					_bytes[p++] = (byte)(B >> _rs0); _bytes[p++] = (byte)(B >> _rs1);
					_bytes[p++] = (byte)(B >> _rs2); _bytes[p++] = (byte)(B >> _rs3);
					B = Float.floatToIntBits(_bufferR[i]);
					_bytes[p++] = (byte)(B >> _rs0); _bytes[p++] = (byte)(B >> _rs1);
					_bytes[p++] = (byte)(B >> _rs2); _bytes[p++] = (byte)(B >> _rs3);
				}
			}
			_sdLine.write(_bytes, 0, _bufferBytes);
			_totalFrames += _frameCnt;
			if (postProcess != null)postProcess.apply();
			if (_stopFlag){ _stopProcess(); return; }
		}
		if (_stopFlag) _stopProcess();
	}
	private void _process_double()
	{
		while ((_dataLineBytes - _sdLine.available()) <= _dlBytes2)
		{
			if (preProcess != null)preProcess.apply();
			if (_pause)for (int i = 0; i < _bufferBytes; i++)_bytes[i] = 0;
			else 
			{
				_clearAudioBuffer();
				_grayblox.audioLatency((_totalFrames - _sdLine.getLongFramePosition()) * _frameTime);
				_grayblox.process(_audioBuffer, 0, _audioBufferSize);
				float[] _bufferL = _audioBuffer[0],  _bufferR = _audioBuffer[1];
				int _rs0 = 0, _rs1 = 8, _rs2 = 16, _rs3 = 24, _rs4 = 32, _rs5 = 40, _rs6 = 48, _rs7 = 56;
				if (_sdLine.getFormat().isBigEndian())
				{_rs0 = 56; _rs1 = 48; _rs2 = 40; _rs3 = 32; _rs4 = 24; _rs5 = 16; _rs6 = 8; _rs7 = 0; }
				for (int i = 0, p = 0; i < _audioBufferSize; i++)
				{
					long B = Double.doubleToLongBits((double)_bufferL[i]);
					_bytes[p++] = (byte)(B >> _rs0); _bytes[p++] = (byte)(B >> _rs1);
					_bytes[p++] = (byte)(B >> _rs2); _bytes[p++] = (byte)(B >> _rs3);
					_bytes[p++] = (byte)(B >> _rs4); _bytes[p++] = (byte)(B >> _rs5);
					_bytes[p++] = (byte)(B >> _rs6); _bytes[p++] = (byte)(B >> _rs7);
					B = Double.doubleToLongBits((double)_bufferR[i]);
					_bytes[p++] = (byte)(B >> _rs0); _bytes[p++] = (byte)(B >> _rs1);
					_bytes[p++] = (byte)(B >> _rs2); _bytes[p++] = (byte)(B >> _rs3);
					_bytes[p++] = (byte)(B >> _rs4); _bytes[p++] = (byte)(B >> _rs5);
					_bytes[p++] = (byte)(B >> _rs6); _bytes[p++] = (byte)(B >> _rs7);
				}
			}
			_sdLine.write(_bytes, 0, _bufferBytes);
			_totalFrames += _frameCnt;
			if (postProcess != null)postProcess.apply();
			if (_stopFlag){ _stopProcess(); return; }
		}
		if (_stopFlag) _stopProcess();
	}
	
	private void _process_s8()
	{
		while ((_dataLineBytes - _sdLine.available()) <= _dlBytes2)
		{
			if (preProcess != null)preProcess.apply();
			if (_pause) for (int i = 0; i < _bufferBytes; i++)_bytes[i] = 0;
			else 
			{
				_clearAudioBuffer();
				_grayblox.audioLatency((_totalFrames - _sdLine.getLongFramePosition()) * _frameTime);
				_grayblox.process(_audioBuffer, 0, _audioBufferSize);
				float[] _bufferL = _audioBuffer[0], _bufferR = _audioBuffer[1];
				for (int i = 0, p = 0; i < _audioBufferSize; i++)
				{
					int _ival = (int)(_bufferL[i] * 127.0f);
					if (_ival > 127)_ival = 127; else if (_ival < -128)_ival = -128;
					_bytes[p++] = (byte)_ival;
					_ival = (int)(_bufferR[i] * 127.0f);
					if (_ival > 127)_ival = 127; else if (_ival < -128)_ival = -128;
					_bytes[p++] = (byte)_ival;
				}
			}
			_sdLine.write(_bytes, 0, _bufferBytes);
			_totalFrames += _frameCnt;
			if (postProcess != null)postProcess.apply();
			if (_stopFlag){ _stopProcess(); return; }
		}
		if (_stopFlag) _stopProcess();
	}
	private void _process_s16()
	{
		while ((_dataLineBytes - _sdLine.available()) <= _dlBytes2)
		{
			if (preProcess != null)preProcess.apply();
			if (_pause) for (int i = 0; i < _bufferBytes; i++)_bytes[i] = 0;
			else 
			{
				_clearAudioBuffer();
				_grayblox.audioLatency((_totalFrames - _sdLine.getLongFramePosition()) * _frameTime);
				_grayblox.process(_audioBuffer, 0, _audioBufferSize);
				float[] _bufferL = _audioBuffer[0], _bufferR = _audioBuffer[1];
				int _rs0 = 0, _rs1 = 8;
				if (_sdLine.getFormat().isBigEndian()){_rs0 = 8; _rs1 = 0; }
				for (int i = 0, p = 0; i < _audioBufferSize; i++)
				{
					int _ival = (int)(_bufferL[i] * 32767.0f);
					if (_ival > 32767)_ival = 32767; else if (_ival < -32768)_ival = -32768;
					_bytes[p++] = (byte)(_ival >> _rs0);
					_bytes[p++] = (byte)(_ival >> _rs1);

					_ival = (int)(_bufferR[i] * 32767.0f);
					if (_ival > 32767)_ival = 32767; else if (_ival < -32768)_ival = -32768;
					_bytes[p++] = (byte)(_ival >> _rs0);
					_bytes[p++] = (byte)(_ival >> _rs1);
				}
			}
			_sdLine.write(_bytes, 0, _bufferBytes);
			_totalFrames += _frameCnt;
			if (postProcess != null)postProcess.apply();
			if (_stopFlag){ _stopProcess(); return; }
		}
		if (_stopFlag) _stopProcess();
	}
	private void _process_s24()
	{
		while ((_dataLineBytes - _sdLine.available()) <= _dlBytes2)
		{
			if (preProcess != null)preProcess.apply();
			if (_pause) for (int i = 0; i < _bufferBytes; i++)_bytes[i] = 0;
			else 
			{
				_clearAudioBuffer();
				_grayblox.audioLatency((_totalFrames - _sdLine.getLongFramePosition()) * _frameTime);
				_grayblox.process(_audioBuffer, 0, _audioBufferSize);
				float[] _bufferL = _audioBuffer[0], _bufferR = _audioBuffer[1];
				int _rs0 = 0, _rs1 = 8, _rs2 = 16;
				if (_sdLine.getFormat().isBigEndian()){_rs0 = 16; _rs1 = 8; _rs1 = 0; }
				for (int i = 0, p = 0; i < _audioBufferSize; i++)
				{
					int _ival = (int)(_bufferL[i] * 8388607.0f);
					if (_ival > 8388607)_ival = 8388607; else if (_ival < -8388608)_ival = -8388608;
					_bytes[p++] = (byte)(_ival >> _rs0);
					_bytes[p++] = (byte)(_ival >> _rs1);
					_bytes[p++] = (byte)(_ival >> _rs2);

					_ival = (int)(_bufferR[i] * 8388607.0f);
					if (_ival > 8388607)_ival = 8388607; else if (_ival < -8388608)_ival = -8388608;
					_bytes[p++] = (byte)(_ival >> _rs0);
					_bytes[p++] = (byte)(_ival >> _rs1);
					_bytes[p++] = (byte)(_ival >> _rs2);
				}
			}
			_sdLine.write(_bytes, 0, _bufferBytes);
			_totalFrames += _frameCnt;
			if (postProcess != null)postProcess.apply();
			if (_stopFlag){ _stopProcess(); return; }
		}
		if (_stopFlag) _stopProcess();
	}
	private void _process_s32()
	{
		while ((_dataLineBytes - _sdLine.available()) <= _dlBytes2)
		{
			if (preProcess != null)preProcess.apply();
			if (_pause) for (int i = 0; i < _bufferBytes; i++)_bytes[i] = 0;
			else 
			{
				_clearAudioBuffer();
				_grayblox.audioLatency((_totalFrames - _sdLine.getLongFramePosition()) * _frameTime);
				_grayblox.process(_audioBuffer, 0, _audioBufferSize);
				float[] _bufferL = _audioBuffer[0], _bufferR = _audioBuffer[1];
				int _rs0 = 0, _rs1 = 8, _rs2 = 16, _rs3 = 24;
				if (_sdLine.getFormat().isBigEndian()){_rs0 = 24; _rs1 = 16; _rs2 = 8; _rs3 = 0; }
				for (int i = 0, p = 0; i < _audioBufferSize; i++)
				{
					long _ival = (long)(_bufferL[i] * 2147483647.0f);
					if (_ival > 2147483647l)_ival = 2147483647l; else if (_ival < -2147483648l)_ival = -2147483648l;
					_bytes[p++] = (byte)(_ival >> _rs0); _bytes[p++] = (byte)(_ival >> _rs1);
					_bytes[p++] = (byte)(_ival >> _rs2); _bytes[p++] = (byte)(_ival >> _rs3);
					_ival = (int)(_bufferR[i] * 2147483647.0f);
					if (_ival > 2147483647l)_ival = 2147483647l; else if (_ival < -2147483648l)_ival = -2147483648l;
					_bytes[p++] = (byte)(_ival >> _rs0); _bytes[p++] = (byte)(_ival >> _rs1);
					_bytes[p++] = (byte)(_ival >> _rs2); _bytes[p++] = (byte)(_ival >> _rs3);
				}
			}
			_sdLine.write(_bytes, 0, _bufferBytes);
			_totalFrames += _frameCnt;
			if (postProcess != null)postProcess.apply();
			if (_stopFlag){ _stopProcess(); return; }
		}
		if (_stopFlag) _stopProcess();
	}

	private void _process_u8()
	{
		while ((_dataLineBytes - _sdLine.available()) <= _dlBytes2)
		{
			if (preProcess != null)preProcess.apply();
			if (_pause) for (int i = 0; i < _bufferBytes; i++)_bytes[i] = 0;
			else 
			{
				_clearAudioBuffer();
				_grayblox.audioLatency((_totalFrames - _sdLine.getLongFramePosition()) * _frameTime);
				_grayblox.process(_audioBuffer, 0, _audioBufferSize);
				float[] _bufferL = _audioBuffer[0], _bufferR = _audioBuffer[1];
				for (int i = 0, p = 0; i < _audioBufferSize; i++)
				{
					int _ival = (int)(_bufferL[i] * 127.0f);
					if (_ival > 127)_ival = 127; else if (_ival < -128)_ival = -128;
					_bytes[p++] = (byte)(_ival + 128);
					_ival = (int)(_bufferR[i] * 127.0f);
					if (_ival > 127)_ival = 127; else if (_ival < -128)_ival = -128;
					_bytes[p++] = (byte)(_ival + 128);
				}
			}
			_sdLine.write(_bytes, 0, _bufferBytes);
			_totalFrames += _frameCnt;
			if (postProcess != null)postProcess.apply();
			if (_stopFlag){ _stopProcess(); return; }
		}
		if (_stopFlag) _stopProcess();
	}
	private void _process_u16()
	{
		while ((_dataLineBytes - _sdLine.available()) <= _dlBytes2)
		{
			if (preProcess != null)preProcess.apply();
			if (_pause) for (int i = 0; i < _bufferBytes; i++)_bytes[i] = 0;
			else 
			{
				_clearAudioBuffer();
				_grayblox.audioLatency((_totalFrames - _sdLine.getLongFramePosition()) * _frameTime);
				_grayblox.process(_audioBuffer, 0, _audioBufferSize);
				float[] _bufferL = _audioBuffer[0], _bufferR = _audioBuffer[1];
				int _rs0 = 0, _rs1 = 8;
				if (_sdLine.getFormat().isBigEndian()){_rs0 = 8; _rs1 = 0; }
				for (int i = 0, p = 0; i < _audioBufferSize; i++)
				{
					int _ival = (int)(_bufferL[i] * 32767.0f);
					if (_ival > 32767)_ival = 32767; else if (_ival < -32768)_ival = -32768;
					_ival += 32768;
					_bytes[p++] = (byte)(_ival >> _rs0);
					_bytes[p++] = (byte)(_ival >> _rs1);

					_ival = (int)(_bufferR[i] * 32767.0f);
					if (_ival > 32767)_ival = 32767; else if (_ival < -32768)_ival = -32768;
					_ival += 32768;
					_bytes[p++] = (byte)(_ival >> _rs0);
					_bytes[p++] = (byte)(_ival >> _rs1);
				}
			}
			_sdLine.write(_bytes, 0, _bufferBytes);
			_totalFrames += _frameCnt;
			if (postProcess != null)postProcess.apply();
			if (_stopFlag){ _stopProcess(); return; }
		}
		if (_stopFlag) _stopProcess();
	}
	private void _process_u24()
	{
		while ((_dataLineBytes - _sdLine.available()) <= _dlBytes2)
		{
			if (preProcess != null)preProcess.apply();
			if (_pause) for (int i = 0; i < _bufferBytes; i++)_bytes[i] = 0;
			else 
			{
				_clearAudioBuffer();
				_grayblox.audioLatency((_totalFrames - _sdLine.getLongFramePosition()) * _frameTime);
				_grayblox.process(_audioBuffer, 0, _audioBufferSize);
				float[] _bufferL = _audioBuffer[0], _bufferR = _audioBuffer[1];
				int _rs0 = 0, _rs1 = 8, _rs2 = 16;
				if (_sdLine.getFormat().isBigEndian()){_rs0 = 16; _rs1 = 8; _rs1 = 0; }
				for (int i = 0, p = 0; i < _audioBufferSize; i++)
				{
					int _ival = (int)(_bufferL[i] * 8388607.0f);
					if (_ival > 8388607)_ival = 8388607; else if (_ival < -8388608)_ival = -8388608;
					_ival += 8388608;
					_bytes[p++] = (byte)(_ival >> _rs0);
					_bytes[p++] = (byte)(_ival >> _rs1);
					_bytes[p++] = (byte)(_ival >> _rs2);

					_ival = (int)(_bufferR[i] * 8388607.0f);
					if (_ival > 8388607)_ival = 8388607; else if (_ival < -8388608)_ival = -8388608;
					_ival += 8388608;
					_bytes[p++] = (byte)(_ival >> _rs0);
					_bytes[p++] = (byte)(_ival >> _rs1);
					_bytes[p++] = (byte)(_ival >> _rs2);
				}
			}
			_sdLine.write(_bytes, 0, _bufferBytes);
			_totalFrames += _frameCnt;
			if (postProcess != null)postProcess.apply();
			if (_stopFlag){ _stopProcess(); return; }
		}
		if (_stopFlag) _stopProcess();
	}
	private void _process_u32()
	{
		while ((_dataLineBytes - _sdLine.available()) <= _dlBytes2)
		{
			if (preProcess != null)preProcess.apply();
			if (_pause) for (int i = 0; i < _bufferBytes; i++)_bytes[i] = 0;
			else 
			{
				_clearAudioBuffer();
				_grayblox.audioLatency((_totalFrames - _sdLine.getLongFramePosition()) * _frameTime);
				_grayblox.process(_audioBuffer, 0, _audioBufferSize);
				float[] _bufferL = _audioBuffer[0], _bufferR = _audioBuffer[1];
				int _rs0 = 0, _rs1 = 8, _rs2 = 16, _rs3 = 24;
				if (_sdLine.getFormat().isBigEndian()){_rs0 = 24; _rs1 = 16; _rs2 = 8; _rs3 = 0; }
				for (int i = 0, p = 0; i < _audioBufferSize; i++)
				{
					long _ival = (long)(_bufferL[i] * 2147483647.0f);
					if (_ival > 2147483647l)_ival = 2147483647l; else if (_ival < -2147483648l)_ival = -2147483648l;
					_ival += 2147483648l;
					_bytes[p++] = (byte)(_ival >> _rs0); _bytes[p++] = (byte)(_ival >> _rs1);
					_bytes[p++] = (byte)(_ival >> _rs2); _bytes[p++] = (byte)(_ival >> _rs3);
					_ival = (int)(_bufferR[i] * 2147483647.0f);
					if (_ival > 2147483647l)_ival = 2147483647l; else if (_ival < -2147483648l)_ival = -2147483648l;
					_ival += 2147483648l;
					_bytes[p++] = (byte)(_ival >> _rs0); _bytes[p++] = (byte)(_ival >> _rs1);
					_bytes[p++] = (byte)(_ival >> _rs2); _bytes[p++] = (byte)(_ival >> _rs3);
				}
			}
			_sdLine.write(_bytes, 0, _bufferBytes);
			_totalFrames += _frameCnt;
			if (postProcess != null)postProcess.apply();
			if (_stopFlag){ _stopProcess(); return; }
		}
		if (_stopFlag) _stopProcess();
	}

private class _BuffTimerTask extends TimerTask
{
	private GrayBoxSound _gbs = null;
	public void run()
	{
		_gbs._process.apply();
	}
	public boolean cancel()
	{
		_gbs = null;
		return super.cancel();
	}
	public _BuffTimerTask(GrayBoxSound gbs)
	{
		_gbs = gbs;
	}
}
}
